#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "nrf_soc.h"

#include "ble_types.h"

#include "GhsControlStructs.h"
#include "nomenclature.h"
#include "configGhsEncoder.h"
#include "msmt_queue.h"

extern unsigned short numberOfStoredMsmtGroups;
extern s_Queue *queue;
extern s_GhsTime *sGhsTime;
extern s_TimeInfo *sTimeInfo;
extern s_TimeInfoData *sTimeInfoData;
extern s_SystemInfoData *systemInfoData;
extern unsigned short msmt_id;
extern unsigned long long epoch;
extern volatile s_global_send global_send;
extern bool first_cont_sent;
extern nrf_mutex_t q_mutex;

bool prepareMeasurements(s_MsmtGroupData *msmtGroupData);

#if (BP_CUFF == 1)
    #define SEND_OPTIMIZED 0
    unsigned short BLE_APPEARANCE = BLE_APPEARANCE_GENERIC_BLOOD_PRESSURE;
    static unsigned char bluetoothAddress[]         = {0x02, 0xF1, 0x02, 0xE3, 0x4D, 0xAA}; // AA:4D:E3:02:F1:01 in little endian
    static unsigned char systemId[]                 = {0x02, 0xF1, 0x02, 0xFF, 0xFF, 0xE3, 0x4D, 0xAA};

    char *DEVICE_NAME                               = "Ghs BP";
    unsigned short SPECIALIZATION                   = MDC_DEV_SPEC_PROFILE_BP;
    static char *MODEL_NUMBER                       = "GHS-BP-234r";
    static char *MANUFACTURER_NAME                  = "GHS Blood Pressure";
    s_MsmtDataBp storedBpMsmts[NUMBER_OF_STORED_MSMTS];    // maximum number of stored msmts NUMBER_OF_STORED_MSMTS.
    s_MsmtGroupData *msmtGroupBpData                = NULL;
    s_MsmtGroupData *msmtGroupOptimizedBpData       = NULL;
    short bp_index                                  = -1;
    short pr_index                                  = -1;
#endif

#if (PULSE_OX == 1)
    #define SEND_OPTIMIZED 1
    unsigned short BLE_APPEARANCE = BLE_APPEARANCE_PULSE_OXIMETER_FINGERTIP;
    static unsigned char bluetoothAddress[]         = {0x01, 0xF1, 0x02, 0xE3, 0x4D, 0xAA}; // AA:4D:E3:02:F1:01 in little endian
    static unsigned char systemId[]                 = {0x01, 0xF1, 0x02, 0xFF, 0xFF, 0xE3, 0x4D, 0xAA};

    char *DEVICE_NAME                               = "Ghs Pulse Ox";
    unsigned short SPECIALIZATION                   = MDC_DEV_SPEC_PROFILE_PULS_OXIM;
    static char *MODEL_NUMBER                       = "GHS-PO-234r";
    static char *MANUFACTURER_NAME                  = "GHS Pulse Oximetry";
    s_MsmtDataPo storedPoMsmts[NUMBER_OF_STORED_MSMTS];    // maximum number of stored msmts NUMBER_OF_STORED_MSMTS.
    s_MsmtGroupData *msmtGroupSpotData              = NULL;
    s_MsmtGroupData *msmtGroupContData              = NULL;
    s_MsmtGroupData *msmtGroupOptimizedContData     = NULL;
    short spo2_index                                = -1;
    short pr_index                                  = -1;
    short qual_index                                = -1;
    short spo2_cont_index                           = -1;
    short pr_cont_index                             = -1;
    short qual_cont_index                           = -1;
#endif
#if (SPIROMETER == 1)
unsigned char flowBytes[] = {
 0x09, 0x00, 0x54, 0x00, 0x8A, 0x01, 0xC9, 0x03, 0x77, 0x05, 0x54, 0x05, 0x7D, 0x04, 0x69, 0x04, 0xD2, 0x04, 0x1D, 0x05, 0x45, 0x05, 0x68, 0x05, 0x77, 0x05, 0x63, 0x05, 0x40, 0x05,
 0x31, 0x05, 0x31, 0x05, 0x36, 0x05, 0x4A, 0x05, 0x3B, 0x05, 0x22, 0x05, 0x2C, 0x05, 0x27, 0x05, 0xFA, 0x04, 0xC3, 0x04, 0xAF, 0x04, 0xBE, 0x04, 0xDC, 0x04, 0xF5, 0x04, 0xF5, 0x04, 0xB9, 0x04, 0x6E, 0x04,
 0x4B, 0x04, 0x64, 0x04, 0xA0, 0x04, 0xC3, 0x04, 0xB4, 0x04, 0xAF, 0x04, 0xB4, 0x04, 0xA0, 0x04, 0x82, 0x04, 0x8C, 0x04, 0xA5, 0x04, 0xB9, 0x04, 0xB9, 0x04, 0xA0, 0x04, 0x8C, 0x04, 0x73, 0x04, 0x46, 0x04,
 0x50, 0x04, 0xA0, 0x04, 0xD7, 0x04, 0xAF, 0x04, 0x55, 0x04, 0x3C, 0x04, 0x5F, 0x04, 0x64, 0x04, 0x50, 0x04, 0x5A, 0x04, 0x55, 0x04, 0x32, 0x04, 0x5F, 0x04, 0x9B, 0x04, 0x8C, 0x04, 0x41, 0x04, 0x28, 0x04,
 0x32, 0x04, 0x3C, 0x04, 0x69, 0x04, 0x82, 0x04, 0x5A, 0x04, 0x32, 0x04, 0x32, 0x04, 0x1E, 0x04, 0xE2, 0x03, 0xEC, 0x03, 0x1E, 0x04, 0xF1, 0x03, 0xC9, 0x03, 0x0A, 0x04, 0x41, 0x04, 0xF1, 0x03, 0x9C, 0x03,
 0xCE, 0x03, 0x00, 0x04, 0xBA, 0x03, 0x97, 0x03, 0xC9, 0x03, 0xBF, 0x03, 0x6A, 0x03, 0x7E, 0x03, 0xF1, 0x03, 0x19, 0x04, 0xD3, 0x03, 0xA1, 0x03, 0x83, 0x03, 0x60, 0x03, 0x88, 0x03, 0xCE, 0x03, 0xA1, 0x03,
 0x6A, 0x03, 0x79, 0x03, 0x92, 0x03, 0x79, 0x03, 0x6A, 0x03, 0x79, 0x03, 0x83, 0x03, 0x8D, 0x03, 0x92, 0x03, 0x6F, 0x03, 0x6F, 0x03, 0x88, 0x03, 0x7E, 0x03, 0x74, 0x03, 0x83, 0x03, 0x65, 0x03, 0x15, 0x03,
 0xFC, 0x02, 0x2E, 0x03, 0x5B, 0x03, 0x65, 0x03, 0x65, 0x03, 0x60, 0x03, 0x3D, 0x03, 0x2E, 0x03, 0x5B, 0x03, 0x79, 0x03, 0x60, 0x03, 0x1A, 0x03, 0x06, 0x03, 0x33, 0x03, 0x3D, 0x03, 0x29, 0x03, 0x3D, 0x03,
 0x24, 0x03, 0xF7, 0x02, 0x38, 0x03, 0x6A, 0x03, 0x15, 0x03, 0xD9, 0x02, 0xF7, 0x02, 0x10, 0x03, 0x29, 0x03, 0x29, 0x03, 0x01, 0x03, 0xCF, 0x02, 0xAC, 0x02, 0xCF, 0x02, 0x15, 0x03, 0x15, 0x03, 0xF2, 0x02,
 0xF2, 0x02, 0xE3, 0x02, 0xB6, 0x02, 0xCA, 0x02, 0xE3, 0x02, 0xC0, 0x02, 0xA7, 0x02, 0xB1, 0x02, 0xD9, 0x02, 0xED, 0x02, 0xDE, 0x02, 0xC0, 0x02, 0xB1, 0x02, 0xB6, 0x02, 0xB6, 0x02, 0xA2, 0x02, 0xB1, 0x02,
 0xC0, 0x02, 0xAC, 0x02, 0xB1, 0x02, 0xC5, 0x02, 0xAC, 0x02, 0xBB, 0x02, 0xE3, 0x02, 0xBB, 0x02, 0x7F, 0x02, 0x8E, 0x02, 0xC5, 0x02, 0xCF, 0x02, 0xAC, 0x02, 0x8E, 0x02, 0x93, 0x02, 0xA2, 0x02, 0x70, 0x02,
 0x61, 0x02, 0x8E, 0x02, 0x70, 0x02, 0x61, 0x02, 0x9D, 0x02, 0xA2, 0x02, 0x89, 0x02, 0x75, 0x02, 0x48, 0x02, 0x3E, 0x02, 0x5C, 0x02, 0x57, 0x02, 0x61, 0x02, 0x93, 0x02, 0x7F, 0x02, 0x2F, 0x02, 0x57, 0x02,
 0x89, 0x02, 0x48, 0x02, 0x6B, 0x02, 0xBB, 0x02, 0x75, 0x02, 0x84, 0x02, 0xD4, 0x02, 0x84, 0x02, 0x2A, 0x02, 0x3E, 0x02, 0x39, 0x02, 0x39, 0x02, 0x70, 0x02, 0x75, 0x02, 0x39, 0x02, 0x34, 0x02, 0x34, 0x02,
 0x20, 0x02, 0x11, 0x02, 0x16, 0x02, 0x34, 0x02, 0x52, 0x02, 0x61, 0x02, 0x6B, 0x02, 0x66, 0x02, 0x4D, 0x02, 0x16, 0x02, 0xEE, 0x01, 0xFD, 0x01, 0x11, 0x02, 0x16, 0x02, 0x39, 0x02, 0x2A, 0x02, 0xD5, 0x01,
 0xDA, 0x01, 0x1B, 0x02, 0x0C, 0x02, 0xFD, 0x01, 0x34, 0x02, 0x43, 0x02, 0x1B, 0x02, 0x0C, 0x02, 0x1B, 0x02, 0x1B, 0x02, 0x25, 0x02, 0x0C, 0x02, 0xF8, 0x01, 0xF3, 0x01, 0xE9, 0x01, 0xE9, 0x01, 0xF8, 0x01,
 0x0C, 0x02, 0x1B, 0x02, 0x07, 0x02, 0xF3, 0x01, 0xF3, 0x01, 0xDA, 0x01, 0xBC, 0x01, 0xE9, 0x01, 0xE9, 0x01, 0xD0, 0x01, 0xF8, 0x01, 0xF8, 0x01, 0xDF, 0x01, 0x2A, 0x02, 0x07, 0x02, 0xB2, 0x01, 0xFD, 0x01,
 0x20, 0x02, 0xE4, 0x01, 0x11, 0x02, 0x0C, 0x02, 0xC1, 0x01, 0xCB, 0x01, 0xCB, 0x01, 0xB2, 0x01, 0xEE, 0x01, 0xE9, 0x01, 0xB2, 0x01, 0xB2, 0x01, 0x94, 0x01, 0x9E, 0x01, 0xD0, 0x01, 0xC6, 0x01, 0xB7, 0x01,
 0xA8, 0x01, 0x9E, 0x01, 0xA8, 0x01, 0xB2, 0x01, 0x9E, 0x01, 0x94, 0x01, 0xB2, 0x01, 0xBC, 0x01, 0xC6, 0x01, 0xC6, 0x01, 0xA8, 0x01, 0xB7, 0x01, 0xC6, 0x01, 0xB7, 0x01, 0xAD, 0x01, 0x6C, 0x01, 0x67, 0x01,
 0xBC, 0x01, 0xB7, 0x01, 0xB7, 0x01, 0xD5, 0x01, 0x8F, 0x01, 0x85, 0x01, 0xE4, 0x01, 0xDA, 0x01, 0xB2, 0x01, 0x9E, 0x01, 0x94, 0x01, 0x9E, 0x01, 0x76, 0x01, 0x76, 0x01, 0xB2, 0x01, 0x8A, 0x01, 0x71, 0x01,
 0xAD, 0x01, 0x8A, 0x01, 0x62, 0x01, 0x71, 0x01, 0x76, 0x01, 0xA8, 0x01, 0x9E, 0x01, 0x5D, 0x01, 0x53, 0x01, 0x4E, 0x01, 0x49, 0x01, 0x8A, 0x01, 0x8A, 0x01, 0x67, 0x01, 0x71, 0x01, 0x62, 0x01, 0x5D, 0x01,
 0x99, 0x01, 0x8A, 0x01, 0x71, 0x01, 0x71, 0x01, 0x44, 0x01, 0x62, 0x01, 0x7B, 0x01, 0x85, 0x01, 0x5D, 0x01, 0x3A, 0x01, 0x76, 0x01, 0x71, 0x01, 0x35, 0x01, 0x53, 0x01, 0x3F, 0x01, 0x21, 0x01, 0x5D, 0x01,
 0x49, 0x01, 0x26, 0x01, 0x4E, 0x01, 0x53, 0x01, 0x4E, 0x01, 0x35, 0x01, 0x30, 0x01, 0x80, 0x01, 0x76, 0x01, 0x3F, 0x01, 0x30, 0x01, 0x53, 0x01, 0x58, 0x01, 0x44, 0x01, 0x30, 0x01, 0x21, 0x01, 0x35, 0x01,
 0x62, 0x01, 0x5D, 0x01, 0x3A, 0x01, 0x2B, 0x01, 0x1C, 0x01, 0x08, 0x01, 0xF9, 0x00, 0x03, 0x01, 0x26, 0x01, 0x35, 0x01, 0x30, 0x01, 0x26, 0x01, 0x30, 0x01, 0x3F, 0x01, 0x26, 0x01, 0x12, 0x01, 0x30, 0x01, 
 0x30, 0x01, 0x35, 0x01, 0x30, 0x01, 0x1C, 0x01, 0xF4, 0x00, 0xF9, 0x00, 0x12, 0x01, 0x17, 0x01, 0x26, 0x01, 0x26, 0x01, 0x17, 0x01, 0x0D, 0x01, 0xD6, 0x00, 0xD6, 0x00, 0x1C, 0x01, 0x35, 0x01, 0x49, 0x01, 
 0x4E, 0x01, 0x35, 0x01, 0x21, 0x01, 0xEA, 0x00, 0xE0, 0x00, 0x21, 0x01, 0x3A, 0x01, 0x3A, 0x01, 0x17, 0x01, 0x03, 0x01, 0xF9, 0x00, 0xEF, 0x00, 0xFE, 0x00, 0x0D, 0x01, 0x1C, 0x01, 0x12, 0x01, 0xD6, 0x00, 
 0xD1, 0x00, 0x0D, 0x01, 0x0D, 0x01, 0xE5, 0x00, 0xE5, 0x00, 0x0D, 0x01, 0x12, 0x01, 0x03, 0x01, 0x0D, 0x01, 0x03, 0x01, 0xD1, 0x00, 0xC2, 0x00, 0xBD, 0x00, 0xD6, 0x00, 0x17, 0x01, 0x12, 0x01, 0xF4, 0x00, 
 0xC7, 0x00, 0xD1, 0x00, 0x12, 0x01, 0xF4, 0x00, 0xDB, 0x00, 0xDB, 0x00, 0xD6, 0x00, 0xC7, 0x00, 0xE5, 0x00, 0xF9, 0x00, 0xEA, 0x00, 0xD6, 0x00, 0xEA, 0x00, 0x0D, 0x01, 0xF9, 0x00, 0xE5, 0x00, 0xE0, 0x00, 
 0xD6, 0x00, 0xC7, 0x00, 0xDB, 0x00, 0xE0, 0x00, 0xD1, 0x00, 0xDB, 0x00, 0xEA, 0x00, 0xC7, 0x00, 0xCC, 0x00, 0xD6, 0x00, 0xB8, 0x00, 0xC2, 0x00, 0xD6, 0x00, 0xEA, 0x00, 0xD1, 0x00, 0xBD, 0x00, 0xB8, 0x00, 
 0xB8, 0x00, 0xE0, 0x00, 0xCC, 0x00, 0xB8, 0x00, 0xDB, 0x00, 0xC2, 0x00, 0xA4, 0x00, 0xA9, 0x00, 0xAE, 0x00, 0xAE, 0x00, 0xD1, 0x00, 0xD6, 0x00, 0xC7, 0x00, 0xE0, 0x00, 0xC7, 0x00, 0xAE, 0x00, 0xDB, 0x00, 
 0xD1, 0x00, 0xBD, 0x00, 0xD1, 0x00, 0xBD, 0x00, 0xBD, 0x00, 0xE5, 0x00, 0xC2, 0x00, 0x9A, 0x00, 0x90, 0x00, 0xB8, 0x00, 0xDB, 0x00, 0xC2, 0x00, 0xAE, 0x00, 0x9F, 0x00, 0xB8, 0x00, 0xB8, 0x00, 0xA9, 0x00, 
 0xA9, 0x00, 0xB3, 0x00, 0xE5, 0x00, 0xD6, 0x00, 0xA9, 0x00, 0x86, 0x00, 0x9A, 0x00, 0xE0, 0x00, 0xE0, 0x00, 0xA9, 0x00, 0x81, 0x00, 0x6D, 0x00, 0x9A, 0x00, 0xEF, 0x00, 0xEA, 0x00, 0xAE, 0x00, 0x81, 0x00, 
 0x86, 0x00, 0x9A, 0x00, 0xC2, 0x00, 0xEF, 0x00, 0xCC, 0x00, 0x95, 0x00, 0x7C, 0x00, 0x77, 0x00, 0x7C, 0x00, 0x86, 0x00, 0xA4, 0x00, 0xB8, 0x00, 0xD6, 0x00, 0xD6, 0x00, 0xB3, 0x00, 0x95, 0x00, 0x77, 0x00, 
 0x68, 0x00, 0x86, 0x00, 0x8B, 0x00, 0x9A, 0x00, 0xA4, 0x00, 0x9F, 0x00, 0x90, 0x00, 0x8B, 0x00, 0x90, 0x00, 0xA4, 0x00, 0x9F, 0x00, 0x95, 0x00, 0x8B, 0x00, 0x90, 0x00, 0x8B, 0x00, 0x86, 0x00, 0x95, 0x00, 
 0x90, 0x00, 0x8B, 0x00, 0x95, 0x00, 0x9F, 0x00, 0x9A, 0x00, 0x90, 0x00, 0x8B, 0x00, 0x90, 0x00, 0x9F, 0x00, 0x9F, 0x00, 0x7C, 0x00, 0x54, 0x00, 0x54, 0x00, 0x63, 0x00, 0x77, 0x00, 0xA9, 0x00, 0xAE, 0x00, 
 0x9A, 0x00, 0x8B, 0x00, 0x90, 0x00, 0x8B, 0x00, 0x72, 0x00, 0x63, 0x00, 0x68, 0x00, 0x81, 0x00, 0x9A, 0x00, 0xAE, 0x00, 0xA4, 0x00, 0x9A, 0x00, 0x77, 0x00, 0x4F, 0x00, 0x40, 0x00, 0x6D, 0x00, 0x9F, 0x00, 
 0xA9, 0x00, 0x9A, 0x00, 0x86, 0x00, 0x63, 0x00, 0x54, 0x00, 0x7C, 0x00, 0x95, 0x00, 0x90, 0x00, 0x68, 0x00, 0x54, 0x00, 0x77, 0x00, 0x81, 0x00, 0x6D, 0x00, 0x4A, 0x00, 0x59, 0x00, 0x77, 0x00, 0x90, 0x00, 
 0x90, 0x00, 0x72, 0x00, 0x72, 0x00, 0x77, 0x00, 0x86, 0x00, 0x7C, 0x00, 0x5E, 0x00, 0x5E, 0x00, 0x77, 0x00, 0x90, 0x00, 0x7C, 0x00, 0x63, 0x00, 0x63, 0x00, 0x5E, 0x00, 0x4A, 0x00, 0x45, 0x00, 0x7C, 0x00, 
 0x95, 0x00, 0x86, 0x00, 0x68, 0x00, 0x4A, 0x00, 0x3B, 0x00, 0x59, 0x00, 0x8B, 0x00, 0xA9, 0x00, 0x86, 0x00, 0x5E, 0x00, 0x59, 0x00, 0x59, 0x00, 0x4F, 0x00, 0x63, 0x00, 0x72, 0x00, 0x68, 0x00, 0x63, 0x00, 
 0x5E, 0x00, 0x40, 0x00, 0x36, 0x00, 0x59, 0x00, 0x6D, 0x00, 0x72, 0x00, 0x90, 0x00, 0x8B, 0x00, 0x68, 0x00, 0x4A, 0x00, 0x3B, 0x00, 0x40, 0x00, 0x72, 0x00, 0x81, 0x00, 0x77, 0x00, 0x81, 0x00, 0x68, 0x00, 
 0x54, 0x00, 0x4F, 0x00, 0x45, 0x00, 0x4A, 0x00, 0x77, 0x00, 0x86, 0x00, 0x63, 0x00, 0x31, 0x00, 0x2C, 0x00, 0x59, 0x00, 0x6D, 0x00, 0x59, 0x00, 0x36, 0x00, 0x4A, 0x00, 0x77, 0x00, 0x81, 0x00, 0x72, 0x00, 
 0x4F, 0x00, 0x40, 0x00, 0x59, 0x00, 0x63, 0x00, 0x4F, 0x00, 0x36, 0x00, 0x54, 0x00, 0x7C, 0x00, 0x81, 0x00, 0x63, 0x00, 0x3B, 0x00, 0x22, 0x00, 0x13, 0x00, 0x36, 0x00, 0x7C, 0x00, 0x90, 0x00, 0x68, 0x00, 
 0x31, 0x00, 0x36, 0x00, 0x40, 0x00, 0x45, 0x00, 0x3B, 0x00, 0x4F, 0x00, 0x68, 0x00, 0x6D, 0x00, 0x6D, 0x00, 0x5E, 0x00, 0x45, 0x00, 0x31, 0x00, 0x40, 0x00, 0x4F, 0x00, 0x45, 0x00, 0x31, 0x00, 0x40, 0x00, 
 0x4F, 0x00, 0x4F, 0x00, 0x45, 0x00, 0x5E, 0x00, 0x6D, 0x00, 0x59, 0x00, 0x31, 0x00, 0x18, 0x00, 0x18, 0x00, 0x45, 0x00, 0x63, 0x00, 0x5E, 0x00, 0x45, 0x00, 0x54, 0x00, 0x63, 0x00, 0x5E, 0x00, 0x40, 0x00, 
 0x40, 0x00, 0x45, 0x00, 0x3B, 0x00, 0x31, 0x00, 0x45, 0x00, 0x59, 0x00, 0x54, 0x00, 0x36, 0x00, 0x2C, 0x00, 0x4F, 0x00, 0x63, 0x00, 0x54, 0x00, 0x36, 0x00, 0x36, 0x00, 0x40, 0x00, 0x40, 0x00, 0x36, 0x00, 
 0x4A, 0x00, 0x59, 0x00, 0x4A, 0x00, 0x31, 0x00, 0x18, 0x00, 0x18, 0x00, 0x36, 0x00, 0x6D, 0x00, 0x77, 0x00, 0x59, 0x00, 0x27, 0x00, 0x13, 0x00, 0x22, 0x00, 0x36, 0x00, 0x3B, 0x00, 0x45, 0x00, 0x6D, 0x00, 
 0x72, 0x00, 0x59, 0x00, 0x27, 0x00, 0x1D, 0x00, 0x31, 0x00, 0x45, 0x00, 0x40, 0x00, 0x22, 0x00, 0x1D, 0x00, 0x3B, 0x00, 0x59, 0x00, 0x54, 0x00, 0x31, 0x00, 0x22, 0x00, 0x45, 0x00, 0x59, 0x00, 0x4A, 0x00, 
 0x27, 0x00, 0x13, 0x00, 0x22, 0x00, 0x31, 0x00, 0x31, 0x00, 0x27, 0x00, 0x1D, 0x00, 0x1D, 0x00, 0x1D, 0x00, 0x31, 0x00, 0x68, 0x00, 0x7C, 0x00, 0x63, 0x00, 0x31, 0x00, 0x18, 0x00, 0x2C, 0x00, 0x3B, 0x00, 
 0x36, 0x00, 0x22, 0x00, 0x13, 0x00, 0x18, 0x00, 0x3B, 0x00, 0x54, 0x00, 0x4A, 0x00, 0x36, 0x00, 0x31, 0x00, 0x59, 0x00, 0x68, 0x00, 0x4F, 0x00, 0x27, 0x00, 0x04, 0x00, 0x04, 0x00, 0x1D, 0x00, 0x36, 0x00, 
 0x3B, 0x00, 0x2C, 0x00, 0x1D, 0x00, 0x22, 0x00, 0x54, 0x00, 0x6D, 0x00, 0x59, 0x00, 0x31, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x09, 0x00, 0x13, 0x00, 0x27, 0x00, 0x4A, 0x00, 0x54, 0x00, 0x3B, 0x00, 0x18, 0x00, 
 0x09, 0x00, 0x09, 0x00, 0x18, 0x00, 0x3B, 0x00, 0x4F, 0x00, 0x45, 0x00, 0x2C, 0x00, 0x1D, 0x00, 0x22, 0x00, 0x4F, 0x00, 0x63, 0x00, 0x4F, 0x00, 0x27, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x0E, 0x00, 0x22, 0x00, 
 0x31, 0x00, 0x2C, 0x00, 0x18, 0x00, 0x0E, 0x00, 0x0E, 0x00, 0x1D, 0x00, 0x54, 0x00, 0x77, 0x00, 0x59, 0x00, 0x27, 0x00, 0x04, 0x00, 0x00, 0x00, 0x13, 0x00, 0x3B, 0x00, 0x45, 0x00, 0x36, 0x00, 0x18, 0x00, 
 0x09, 0x00, 0x09, 0x00, 0x09, 0x00, 0x0E, 0x00, 0x31, 0x00, 0x40, 0x00, 0x3B, 0x00, 0x27, 0x00, 0x18, 0x00, 0x09, 0x00, 0x09, 0x00, 0x13, 0x00, 0x1D, 0x00, 0x1D, 0x00, 0x27, 0x00, 0x54, 0x00, 0x68, 0x00, 
 0x4F, 0x00, 0x1D, 0x00, 0x04, 0x00, 0x00, 0x00, 0x09, 0x00, 0x18, 0x00, 0x2C, 0x00, 0x3B, 0x00, 0x2C, 0x00, 0x1D, 0x00, 0x13, 0x00, 0x13, 0x00, 0x13, 0x00, 0x1D, 0x00, 0x4F, 0x00, 0x6D, 0x00, 0x59, 0x00, 
 0x31, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x04, 0x00, 0x13, 0x00, 0x31, 0x00, 0x45, 0x00, 0x36, 0x00, 0x18, 0x00, 0x09, 0x00, 0x09, 0x00, 0x0E, 0x00, 0x13, 0x00, 0x18, 0x00, 0x31, 0x00, 0x45, 0x00, 0x36, 0x00, 
 0x1D, 0x00, 0x09, 0x00, 0x09, 0x00, 0x09, 0x00, 0x0E, 0x00, 0x1D, 0x00, 0x4F, 0x00, 0x63, 0x00, 0x4F, 0x00, 0x1D, 0x00, 0xFC, 0xFF, 0xF7, 0xFF, 0xFC, 0xFF, 0x00, 0x00, 0x09, 0x00, 0x0E, 0x00, 0x22, 0x00, 
 0x31, 0x00, 0x2C, 0x00, 0x1D, 0x00, 0x13, 0x00, 0x13, 0x00, 0x0E, 0x00, 0x09, 0x00, 0x09, 0x00, 0x13, 0x00, 0x13, 0x00, 0x13, 0x00, 0x13, 0x00, 0x09, 0x00, 0x18, 0x00, 0x40, 0x00, 0x5E, 0x00, 0x4F, 0x00, 
 0x27, 0x00, 0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 0x09, 0x00, 0x09, 0x00, 0x18, 0x00, 0x2C, 0x00, 0x31, 0x00, 0x2C, 0x00, 0x18, 0x00, 0x04, 0x00, 0x04, 0x00, 0x0E, 0x00, 0x13, 0x00, 0x13, 0x00, 0x0E, 0x00, 
 0x0E, 0x00, 0x27, 0x00, 0x3B, 0x00, 0x2C, 0x00, 0x18, 0x00, 0x04, 0x00, 0x00, 0x00, 0x09, 0x00, 0x13, 0x00, 0x13, 0x00, 0x0E, 0x00, 0x09, 0x00, 0x13, 0x00, 0x40, 0x00, 0x63, 0x00, 0x4F, 0x00, 0x1D, 0x00, 
 0xFC, 0xFF, 0xF7, 0xFF, 0x04, 0x00, 0x0E, 0x00, 0x13, 0x00, 0x09, 0x00, 0x00, 0x00, 0x09, 0x00, 0x22, 0x00, 0x31, 0x00, 0x2C, 0x00, 0x18, 0x00, 0x09, 0x00, 0x00, 0x00, 0x04, 0x00, 0x09, 0x00};
/*
unsigned char volBytes[] = {0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x0E, 0x00, 0x1C, 0x00, 0x2A, 0x00, 0x35, 0x00, 0x40, 0x00, 0x4D, 0x00, 0x5A, 0x00, 0x67, 0x00, 0x75, 0x00, 0x83, 0x00, 0x91, 0x00, 0x9E, 0x00,
 0xAC, 0x00, 0xB9, 0x00, 0xC6, 0x00, 0xD4, 0x00, 0xE1, 0x00, 0xEF, 0x00, 0xFC, 0x00, 0x09, 0x01, 0x16, 0x01, 0x22, 0x01, 0x2E, 0x01, 0x3A, 0x01, 0x47, 0x01, 0x53, 0x01, 0x60, 0x01, 0x6C, 0x01, 0x77, 0x01, 
 0x82, 0x01, 0x8E, 0x01, 0x99, 0x01, 0xA6, 0x01, 0xB2, 0x01, 0xBE, 0x01, 0xCA, 0x01, 0xD6, 0x01, 0xE1, 0x01, 0xED, 0x01, 0xF9, 0x01, 0x05, 0x02, 0x11, 0x02, 0x1D, 0x02, 0x28, 0x02, 0x34, 0x02, 0x3F, 0x02, 
 0x4A, 0x02, 0x56, 0x02, 0x62, 0x02, 0x6E, 0x02, 0x79, 0x02, 0x84, 0x02, 0x8F, 0x02, 0x9A, 0x02, 0xA6, 0x02, 0xB1, 0x02, 0xBC, 0x02, 0xC7, 0x02, 0xD2, 0x02, 0xDE, 0x02, 0xE9, 0x02, 0xF4, 0x02, 0xFF, 0x02, 
 0x09, 0x03, 0x14, 0x03, 0x20, 0x03, 0x2B, 0x03, 0x36, 0x03, 0x41, 0x03, 0x4C, 0x03, 0x56, 0x03, 0x60, 0x03, 0x6A, 0x03, 0x75, 0x03, 0x7F, 0x03, 0x89, 0x03, 0x93, 0x03, 0x9E, 0x03, 0xA8, 0x03, 0xB1, 0x03, 
 0xBB, 0x03, 0xC5, 0x03, 0xCF, 0x03, 0xD8, 0x03, 0xE2, 0x03, 0xEB, 0x03, 0xF4, 0x03, 0xFD, 0x03, 0x07, 0x04, 0x12, 0x04, 0x1B, 0x04, 0x25, 0x04, 0x2E, 0x04, 0x36, 0x04, 0x3F, 0x04, 0x49, 0x04, 0x53, 0x04, 
 0x5B, 0x04, 0x64, 0x04, 0x6D, 0x04, 0x76, 0x04, 0x7F, 0x04, 0x88, 0x04, 0x91, 0x04, 0x9A, 0x04, 0xA3, 0x04, 0xAC, 0x04, 0xB5, 0x04, 0xBE, 0x04, 0xC7, 0x04, 0xD0, 0x04, 0xD9, 0x04, 0xE1, 0x04, 0xE9, 0x04, 
 0xF1, 0x04, 0xF9, 0x04, 0x02, 0x05, 0x0A, 0x05, 0x13, 0x05, 0x1C, 0x05, 0x24, 0x05, 0x2C, 0x05, 0x35, 0x05, 0x3E, 0x05, 0x46, 0x05, 0x4E, 0x05, 0x56, 0x05, 0x5E, 0x05, 0x66, 0x05, 0x6F, 0x05, 0x77, 0x05, 
 0x7F, 0x05, 0x86, 0x05, 0x8F, 0x05, 0x97, 0x05, 0x9F, 0x05, 0xA7, 0x05, 0xAE, 0x05, 0xB6, 0x05, 0xBE, 0x05, 0xC6, 0x05, 0xCE, 0x05, 0xD5, 0x05, 0xDC, 0x05, 0xE3, 0x05, 0xEB, 0x05, 0xF3, 0x05, 0xFB, 0x05, 
 0x02, 0x06, 0x0A, 0x06, 0x11, 0x06, 0x18, 0x06, 0x1F, 0x06, 0x26, 0x06, 0x2D, 0x06, 0x34, 0x06, 0x3B, 0x06, 0x43, 0x06, 0x4A, 0x06, 0x51, 0x06, 0x58, 0x06, 0x5F, 0x06, 0x66, 0x06, 0x6D, 0x06, 0x73, 0x06, 
 0x7B, 0x06, 0x81, 0x06, 0x88, 0x06, 0x8F, 0x06, 0x96, 0x06, 0x9D, 0x06, 0xA5, 0x06, 0xAC, 0x06, 0xB2, 0x06, 0xB9, 0x06, 0xC0, 0x06, 0xC7, 0x06, 0xCE, 0x06, 0xD4, 0x06, 0xDB, 0x06, 0xE2, 0x06, 0xE8, 0x06, 
 0xEE, 0x06, 0xF5, 0x06, 0xFB, 0x06, 0x01, 0x07, 0x08, 0x07, 0x0E, 0x07, 0x15, 0x07, 0x1B, 0x07, 0x21, 0x07, 0x27, 0x07, 0x2D, 0x07, 0x33, 0x07, 0x39, 0x07, 0x3F, 0x07, 0x46, 0x07, 0x4B, 0x07, 0x51, 0x07, 
 0x58, 0x07, 0x5E, 0x07, 0x64, 0x07, 0x6B, 0x07, 0x71, 0x07, 0x78, 0x07, 0x7F, 0x07, 0x85, 0x07, 0x8B, 0x07, 0x91, 0x07, 0x96, 0x07, 0x9C, 0x07, 0xA2, 0x07, 0xA9, 0x07, 0xAE, 0x07, 0xB4, 0x07, 0xBA, 0x07, 
 0xBF, 0x07, 0xC4, 0x07, 0xCA, 0x07, 0xCF, 0x07, 0xD5, 0x07, 0xDC, 0x07, 0xE2, 0x07, 0xE8, 0x07, 0xEE, 0x07, 0xF3, 0x07, 0xF8, 0x07, 0xFD, 0x07, 0x02, 0x08, 0x08, 0x08, 0x0E, 0x08, 0x13, 0x08, 0x18, 0x08, 
 0x1D, 0x08, 0x22, 0x08, 0x27, 0x08, 0x2C, 0x08, 0x32, 0x08, 0x38, 0x08, 0x3D, 0x08, 0x42, 0x08, 0x48, 0x08, 0x4D, 0x08, 0x53, 0x08, 0x58, 0x08, 0x5D, 0x08, 0x62, 0x08, 0x67, 0x08, 0x6C, 0x08, 0x71, 0x08, 
 0x76, 0x08, 0x7B, 0x08, 0x81, 0x08, 0x86, 0x08, 0x8B, 0x08, 0x8F, 0x08, 0x94, 0x08, 0x99, 0x08, 0x9E, 0x08, 0xA2, 0x08, 0xA7, 0x08, 0xAC, 0x08, 0xB1, 0x08, 0xB7, 0x08, 0xBC, 0x08, 0xC0, 0x08, 0xC5, 0x08, 
 0xCB, 0x08, 0xD0, 0x08, 0xD5, 0x08, 0xDA, 0x08, 0xDF, 0x08, 0xE3, 0x08, 0xE8, 0x08, 0xEC, 0x08, 0xF1, 0x08, 0xF6, 0x08, 0xFB, 0x08, 0xFF, 0x08, 0x03, 0x09, 0x07, 0x09, 0x0C, 0x09, 0x10, 0x09, 0x15, 0x09, 
 0x19, 0x09, 0x1D, 0x09, 0x21, 0x09, 0x26, 0x09, 0x2A, 0x09, 0x2E, 0x09, 0x32, 0x09, 0x37, 0x09, 0x3B, 0x09, 0x40, 0x09, 0x44, 0x09, 0x48, 0x09, 0x4D, 0x09, 0x51, 0x09, 0x56, 0x09, 0x59, 0x09, 0x5D, 0x09, 
 0x61, 0x09, 0x66, 0x09, 0x6A, 0x09, 0x6F, 0x09, 0x73, 0x09, 0x77, 0x09, 0x7C, 0x09, 0x80, 0x09, 0x85, 0x09, 0x89, 0x09, 0x8D, 0x09, 0x91, 0x09, 0x95, 0x09, 0x99, 0x09, 0x9D, 0x09, 0xA1, 0x09, 0xA5, 0x09, 
 0xA9, 0x09, 0xAD, 0x09, 0xB0, 0x09, 0xB4, 0x09, 0xB8, 0x09, 0xBC, 0x09, 0xC0, 0x09, 0xC4, 0x09, 0xC7, 0x09, 0xCA, 0x09, 0xCE, 0x09, 0xD2, 0x09, 0xD6, 0x09, 0xD9, 0x09, 0xDD, 0x09, 0xE1, 0x09, 0xE4, 0x09, 
 0xE8, 0x09, 0xEC, 0x09, 0xF0, 0x09, 0xF3, 0x09, 0xF7, 0x09, 0xFA, 0x09, 0xFE, 0x09, 0x02, 0x0A, 0x05, 0x0A, 0x09, 0x0A, 0x0C, 0x0A, 0x10, 0x0A, 0x13, 0x0A, 0x17, 0x0A, 0x1A, 0x0A, 0x1D, 0x0A, 0x20, 0x0A, 
 0x23, 0x0A, 0x26, 0x0A, 0x2A, 0x0A, 0x2D, 0x0A, 0x31, 0x0A, 0x34, 0x0A, 0x37, 0x0A, 0x3B, 0x0A, 0x3E, 0x0A, 0x41, 0x0A, 0x45, 0x0A, 0x48, 0x0A, 0x4B, 0x0A, 0x4F, 0x0A, 0x52, 0x0A, 0x55, 0x0A, 0x58, 0x0A, 
 0x5B, 0x0A, 0x5F, 0x0A, 0x62, 0x0A, 0x65, 0x0A, 0x68, 0x0A, 0x6A, 0x0A, 0x6D, 0x0A, 0x6F, 0x0A, 0x72, 0x0A, 0x76, 0x0A, 0x79, 0x0A, 0x7C, 0x0A, 0x7F, 0x0A, 0x82, 0x0A, 0x85, 0x0A, 0x87, 0x0A, 0x8B, 0x0A, 
 0x8E, 0x0A, 0x91, 0x0A, 0x94, 0x0A, 0x97, 0x0A, 0x99, 0x0A, 0x9C, 0x0A, 0x9E, 0x0A, 0xA1, 0x0A, 0xA4, 0x0A, 0xA7, 0x0A, 0xAA, 0x0A, 0xAC, 0x0A, 0xAF, 0x0A, 0xB1, 0x0A, 0xB4, 0x0A, 0xB7, 0x0A, 0xBA, 0x0A, 
 0xBD, 0x0A, 0xC0, 0x0A, 0xC3, 0x0A, 0xC6, 0x0A, 0xC8, 0x0A, 0xCB, 0x0A, 0xCE, 0x0A, 0xD1, 0x0A, 0xD4, 0x0A, 0xD7, 0x0A, 0xD9, 0x0A, 0xDB, 0x0A, 0xDE, 0x0A, 0xE1, 0x0A, 0xE4, 0x0A, 0xE6, 0x0A, 0xE8, 0x0A, 
 0xEB, 0x0A, 0xED, 0x0A, 0xF0, 0x0A, 0xF2, 0x0A, 0xF5, 0x0A, 0xF7, 0x0A, 0xFA, 0x0A, 0xFD, 0x0A, 0xFF, 0x0A, 0x02, 0x0B, 0x04, 0x0B, 0x06, 0x0B, 0x08, 0x0B, 0x0A, 0x0B, 0x0D, 0x0B, 0x10, 0x0B, 0x12, 0x0B, 
 0x14, 0x0B, 0x16, 0x0B, 0x19, 0x0B, 0x1B, 0x0B, 0x1E, 0x0B, 0x20, 0x0B, 0x22, 0x0B, 0x24, 0x0B, 0x26, 0x0B, 0x29, 0x0B, 0x2B, 0x0B, 0x2D, 0x0B, 0x30, 0x0B, 0x32, 0x0B, 0x35, 0x0B, 0x37, 0x0B, 0x39, 0x0B, 
 0x3B, 0x0B, 0x3D, 0x0B, 0x40, 0x0B, 0x42, 0x0B, 0x44, 0x0B, 0x46, 0x0B, 0x49, 0x0B, 0x4B, 0x0B, 0x4D, 0x0B, 0x4F, 0x0B, 0x51, 0x0B, 0x53, 0x0B, 0x55, 0x0B, 0x57, 0x0B, 0x59, 0x0B, 0x5B, 0x0B, 0x5D, 0x0B, 
 0x5F, 0x0B, 0x61, 0x0B, 0x63, 0x0B, 0x65, 0x0B, 0x67, 0x0B, 0x69, 0x0B, 0x6B, 0x0B, 0x6C, 0x0B, 0x6E, 0x0B, 0x70, 0x0B, 0x72, 0x0B, 0x74, 0x0B, 0x76, 0x0B, 0x78, 0x0B, 0x7A, 0x0B, 0x7C, 0x0B, 0x7E, 0x0B, 
 0x80, 0x0B, 0x82, 0x0B, 0x84, 0x0B, 0x86, 0x0B, 0x88, 0x0B, 0x8B, 0x0B, 0x8C, 0x0B, 0x8E, 0x0B, 0x8F, 0x0B, 0x91, 0x0B, 0x94, 0x0B, 0x95, 0x0B, 0x97, 0x0B, 0x99, 0x0B, 0x9B, 0x0B, 0x9D, 0x0B, 0x9E, 0x0B, 
 0xA0, 0x0B, 0xA2, 0x0B, 0xA4, 0x0B, 0xA6, 0x0B, 0xA8, 0x0B, 0xA9, 0x0B, 0xAB, 0x0B, 0xAD, 0x0B, 0xAF, 0x0B, 0xB1, 0x0B, 0xB2, 0x0B, 0xB3, 0x0B, 0xB5, 0x0B, 0xB7, 0x0B, 0xBA, 0x0B, 0xBB, 0x0B, 0xBD, 0x0B, 
 0xBE, 0x0B, 0xC0, 0x0B, 0xC2, 0x0B, 0xC4, 0x0B, 0xC6, 0x0B, 0xC8, 0x0B, 0xC9, 0x0B, 0xCA, 0x0B, 0xCB, 0x0B, 0xCD, 0x0B, 0xCE, 0x0B, 0xD0, 0x0B, 0xD2, 0x0B, 0xD4, 0x0B, 0xD6, 0x0B, 0xD8, 0x0B, 0xD9, 0x0B, 
 0xDA, 0x0B, 0xDB, 0x0B, 0xDD, 0x0B, 0xDE, 0x0B, 0xE0, 0x0B, 0xE1, 0x0B, 0xE3, 0x0B, 0xE4, 0x0B, 0xE6, 0x0B, 0xE7, 0x0B, 0xE9, 0x0B, 0xEB, 0x0B, 0xEC, 0x0B, 0xED, 0x0B, 0xEF, 0x0B, 0xF0, 0x0B, 0xF2, 0x0B, 
 0xF3, 0x0B, 0xF4, 0x0B, 0xF6, 0x0B, 0xF8, 0x0B, 0xF9, 0x0B, 0xFB, 0x0B, 0xFC, 0x0B, 0xFD, 0x0B, 0xFF, 0x0B, 0x01, 0x0C, 0x02, 0x0C, 0x03, 0x0C, 0x04, 0x0C, 0x05, 0x0C, 0x06, 0x0C, 0x07, 0x0C, 0x09, 0x0C, 
 0x0B, 0x0C, 0x0C, 0x0C, 0x0E, 0x0C, 0x0F, 0x0C, 0x10, 0x0C, 0x11, 0x0C, 0x12, 0x0C, 0x14, 0x0C, 0x15, 0x0C, 0x17, 0x0C, 0x18, 0x0C, 0x1A, 0x0C, 0x1B, 0x0C, 0x1C, 0x0C, 0x1D, 0x0C, 0x1E, 0x0C, 0x1F, 0x0C, 
 0x21, 0x0C, 0x23, 0x0C, 0x24, 0x0C, 0x25, 0x0C, 0x26, 0x0C, 0x27, 0x0C, 0x29, 0x0C, 0x2A, 0x0C, 0x2B, 0x0C, 0x2C, 0x0C, 0x2D, 0x0C, 0x2E, 0x0C, 0x30, 0x0C, 0x30, 0x0C, 0x31, 0x0C, 0x32, 0x0C, 0x34, 0x0C, 
 0x35, 0x0C, 0x36, 0x0C, 0x38, 0x0C, 0x39, 0x0C, 0x3A, 0x0C, 0x3B, 0x0C, 0x3C, 0x0C, 0x3D, 0x0C, 0x3E, 0x0C, 0x40, 0x0C, 0x41, 0x0C, 0x42, 0x0C, 0x43, 0x0C, 0x44, 0x0C, 0x45, 0x0C, 0x46, 0x0C, 0x47, 0x0C, 
 0x48, 0x0C, 0x4A, 0x0C, 0x4B, 0x0C, 0x4B, 0x0C, 0x4C, 0x0C, 0x4D, 0x0C, 0x4E, 0x0C, 0x50, 0x0C, 0x51, 0x0C, 0x52, 0x0C, 0x53, 0x0C, 0x54, 0x0C, 0x55, 0x0C, 0x56, 0x0C, 0x57, 0x0C, 0x58, 0x0C, 0x59, 0x0C, 
 0x5A, 0x0C, 0x5B, 0x0C, 0x5B, 0x0C, 0x5C, 0x0C, 0x5D, 0x0C, 0x5E, 0x0C, 0x60, 0x0C, 0x61, 0x0C, 0x62, 0x0C, 0x63, 0x0C, 0x64, 0x0C, 0x64, 0x0C, 0x66, 0x0C, 0x67, 0x0C, 0x68, 0x0C, 0x69, 0x0C, 0x6A, 0x0C, 
 0x6B, 0x0C, 0x6C, 0x0C, 0x6D, 0x0C, 0x6D, 0x0C, 0x6F, 0x0C, 0x70, 0x0C, 0x71, 0x0C, 0x72, 0x0C, 0x72, 0x0C, 0x73, 0x0C, 0x74, 0x0C, 0x75, 0x0C, 0x75, 0x0C, 0x76, 0x0C, 0x77, 0x0C, 0x79, 0x0C, 0x7A, 0x0C, 
 0x7B, 0x0C, 0x7B, 0x0C, 0x7C, 0x0C, 0x7D, 0x0C, 0x7E, 0x0C, 0x7F, 0x0C, 0x7F, 0x0C, 0x81, 0x0C, 0x82, 0x0C, 0x83, 0x0C, 0x84, 0x0C, 0x84, 0x0C, 0x84, 0x0C, 0x85, 0x0C, 0x86, 0x0C, 0x87, 0x0C, 0x88, 0x0C, 
 0x89, 0x0C, 0x89, 0x0C, 0x8A, 0x0C, 0x8B, 0x0C, 0x8B, 0x0C, 0x8C, 0x0C, 0x8D, 0x0C, 0x8E, 0x0C, 0x8F, 0x0C, 0x90, 0x0C, 0x91, 0x0C, 0x92, 0x0C, 0x92, 0x0C, 0x93, 0x0C, 0x94, 0x0C, 0x94, 0x0C, 0x95, 0x0C, 
 0x96, 0x0C, 0x96, 0x0C, 0x97, 0x0C, 0x98, 0x0C, 0x99, 0x0C, 0x9A, 0x0C, 0x9B, 0x0C, 0x9B, 0x0C, 0x9B, 0x0C, 0x9C, 0x0C, 0x9D, 0x0C, 0x9E, 0x0C, 0x9E, 0x0C, 0x9F, 0x0C, 0xA0, 0x0C, 0xA1, 0x0C, 0xA2, 0x0C, 
 0xA3, 0x0C, 0xA3, 0x0C, 0xA4, 0x0C, 0xA4, 0x0C, 0xA5, 0x0C, 0xA6, 0x0C, 0xA7, 0x0C, 0xA7, 0x0C, 0xA8, 0x0C, 0xA9, 0x0C, 0xAA, 0x0C, 0xAA, 0x0C, 0xAB, 0x0C, 0xAC, 0x0C, 0xAC, 0x0C, 0xAD, 0x0C, 0xAD, 0x0C, 
 0xAE, 0x0C, 0xAF, 0x0C, 0xB0, 0x0C, 0xB0, 0x0C, 0xB1, 0x0C, 0xB1, 0x0C, 0xB1, 0x0C, 0xB2, 0x0C, 0xB4, 0x0C, 0xB5, 0x0C, 0xB5, 0x0C, 0xB5, 0x0C, 0xB6, 0x0C, 0xB6, 0x0C, 0xB7, 0x0C, 0xB7, 0x0C, 0xB8, 0x0C, 
 0xBA, 0x0C, 0xBB, 0x0C, 0xBB, 0x0C, 0xBB, 0x0C, 0xBC, 0x0C, 0xBC, 0x0C, 0xBD, 0x0C, 0xBD, 0x0C, 0xBE, 0x0C, 0xBE, 0x0C, 0xBF, 0x0C, 0xC0, 0x0C, 0xC1, 0x0C, 0xC1, 0x0C, 0xC2, 0x0C, 0xC3, 0x0C, 0xC3, 0x0C, 
 0xC4, 0x0C, 0xC4, 0x0C, 0xC4, 0x0C, 0xC5, 0x0C, 0xC5, 0x0C, 0xC6, 0x0C, 0xC6, 0x0C, 0xC6, 0x0C, 0xC7, 0x0C, 0xC7, 0x0C, 0xC8, 0x0C, 0xC9, 0x0C, 0xCA, 0x0C, 0xCB, 0x0C, 0xCB, 0x0C, 0xCC, 0x0C, 0xCC, 0x0C, 
 0xCD, 0x0C, 0xCD, 0x0C, 0xCD, 0x0C, 0xCD, 0x0C, 0xCE, 0x0C, 0xCF, 0x0C, 0xD0, 0x0C, 0xD0, 0x0C, 0xD1, 0x0C, 0xD2, 0x0C, 0xD3, 0x0C, 0xD3, 0x0C, 0xD4, 0x0C, 0xD4, 0x0C, 0xD4, 0x0C, 0xD4, 0x0C, 0xD5, 0x0C, 
 0xD5, 0x0C, 0xD6, 0x0C, 0xD6, 0x0C, 0xD7, 0x0C, 0xD7, 0x0C, 0xD8, 0x0C, 0xD9, 0x0C, 0xDA, 0x0C, 0xDA, 0x0C, 0xDA, 0x0C, 0xDA, 0x0C, 0xDA, 0x0C, 0xDB, 0x0C, 0xDB, 0x0C, 0xDC, 0x0C, 0xDD, 0x0C, 0xDD, 0x0C, 
 0xDD, 0x0C, 0xDD, 0x0C, 0xDE, 0x0C, 0xDE, 0x0C, 0xDF, 0x0C, 0xE0, 0x0C, 0xE0, 0x0C, 0xE0, 0x0C, 0xE1, 0x0C, 0xE2, 0x0C, 0xE3, 0x0C, 0xE3, 0x0C, 0xE4, 0x0C, 0xE4, 0x0C, 0xE4, 0x0C, 0xE4, 0x0C, 0xE4, 0x0C, 
 0xE5, 0x0C, 0xE5, 0x0C, 0xE6, 0x0C, 0xE6, 0x0C, 0xE6, 0x0C, 0xE6, 0x0C, 0xE7, 0x0C, 0xE8, 0x0C, 0xE9, 0x0C, 0xEA, 0x0C, 0xEA, 0x0C, 0xEA, 0x0C, 0xEA, 0x0C, 0xEA, 0x0C, 0xEB, 0x0C, 0xEC, 0x0C, 0xEC, 0x0C, 
 0xEC, 0x0C, 0xEC, 0x0C, 0xEC, 0x0C, 0xEC, 0x0C, 0xED, 0x0C, 0xEE, 0x0C, 0xEE, 0x0C, 0xEF, 0x0C, 0xEF, 0x0C, 0xEF, 0x0C, 0xEF, 0x0C, 0xEF, 0x0C, 0xF0, 0x0C, 0xF0, 0x0C, 0xF0, 0x0C, 0xF1, 0x0C, 0xF2, 0x0C, 
 0xF3, 0x0C, 0xF3, 0x0C, 0xF3, 0x0C, 0xF3, 0x0C, 0xF3, 0x0C, 0xF4, 0x0C, 0xF4, 0x0C, 0xF5, 0x0C, 0xF5, 0x0C, 0xF5, 0x0C, 0xF6, 0x0C, 0xF6, 0x0C, 0xF6, 0x0C, 0xF6, 0x0C, 0xF7, 0x0C, 0xF8, 0x0C, 0xF9, 0x0C, 
 0xFA, 0x0C, 0xFA, 0x0C, 0xFA, 0x0C, 0xFA, 0x0C, 0xFA, 0x0C, 0xFB, 0x0C, 0xFB, 0x0C, 0xFC, 0x0C, 0xFC, 0x0C, 0xFC, 0x0C, 0xFC, 0x0C, 0xFC, 0x0C, 0xFD, 0x0C, 0xFD, 0x0C, 0xFD, 0x0C, 0xFE, 0x0C, 0xFF, 0x0C, 
 0xFF, 0x0C, 0xFF, 0x0C, 0xFF, 0x0C, 0xFF, 0x0C, 0xFF, 0x0C, 0x00, 0x0D, 0x00, 0x0D, 0x01, 0x0D, 0x02, 0x0D, 0x03, 0x0D, 0x02, 0x0D, 0x02, 0x0D, 0x02, 0x0D, 0x02, 0x0D, 0x02, 0x0D, 0x03, 0x0D, 0x03, 0x0D, 
 0x03, 0x0D, 0x04, 0x0D, 0x04, 0x0D, 0x04, 0x0D, 0x05, 0x0D, 0x05, 0x0D, 0x05, 0x0D, 0x05, 0x0D, 0x05, 0x0D, 0x05, 0x0D, 0x06, 0x0D, 0x06, 0x0D, 0x06, 0x0D, 0x06, 0x0D, 0x07, 0x0D, 0x08, 0x0D, 0x08, 0x0D, 
 0x09, 0x0D, 0x09, 0x0D, 0x09, 0x0D, 0x09, 0x0D, 0x09, 0x0D, 0x09, 0x0D, 0x09, 0x0D, 0x0A, 0x0D, 0x0A, 0x0D, 0x0B, 0x0D, 0x0B, 0x0D, 0x0B, 0x0D, 0x0B, 0x0D, 0x0B, 0x0D, 0x0C, 0x0D, 0x0C, 0x0D, 0x0C, 0x0D, 
 0x0C, 0x0D, 0x0C, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0E, 0x0D, 0x0E, 0x0D, 0x0E, 0x0D, 0x0E, 0x0D, 0x0E, 0x0D, 0x0E, 0x0D, 0x0E, 0x0D, 0x0F, 0x0D, 0x0F, 0x0D, 0x0F, 0x0D, 0x10, 0x0D, 0x11, 0x0D, 0x11, 0x0D, 
 0x11, 0x0D, 0x11, 0x0D, 0x11, 0x0D, 0x12, 0x0D, 0x12, 0x0D, 0x12, 0x0D, 0x12, 0x0D, 0x12, 0x0D, 0x12, 0x0D, 0x13, 0x0D, 0x13, 0x0D, 0x13, 0x0D, 0x14, 0x0D, 0x14, 0x0D, 0x14, 0x0D, 0x14, 0x0D};
 */
    unsigned short fev1Val = 1137;
    unsigned short fev6Val = 3117;
    unsigned short fvcVal  = 3393;
    unsigned short pefVal  = 1450;
    unsigned short fetVal  = 10000;
    short fev1RatingsVal[3] = {-1500, 2681, 324};      // Z score 8 1000, LLN * 1000, % predicted * 10
    unsigned long fvcAtsGradeVal = MDC_SPIRO_AST_QUAL_D;
    unsigned long fev1AtsGradeVal = MDC_SPIRO_AST_QUAL_D;

    #define SEND_OPTIMIZED 0
    #define AGE 71
    #define HEIGHT 1829
    #define WEIGHT 9072
    #define NO_OF_SAMPLES 500
    #define SAMPLE_SIZE 2
    unsigned short BLE_APPEARANCE = BLE_APPEARANCE_UNKNOWN;
    static unsigned char bluetoothAddress[]         = {0x03, 0xF1, 0x02, 0xE3, 0x4D, 0xAA}; // AA:4D:E3:02:F1:01 in little endian
    static unsigned char systemId[]                 = {0x03, 0xF1, 0x02, 0xFF, 0xFF, 0xE3, 0x4D, 0xAA};

    char *DEVICE_NAME                               = "Ghs Spirometer";
    unsigned short SPECIALIZATION                   = MDC_DEV_SPEC_PROFILE_SPIRO;
    static char *MODEL_NUMBER                       = "GHS-SPIRO-234r";
    static char *MANUFACTURER_NAME                  = "GHS Spirometry";

    s_MsmtGroupData *msmtGroupSpiroSettingsData     = NULL;
    s_MsmtGroupData *msmtGroupSpiroManeuvData       = NULL;
    s_MsmtGroupData *msmtGroupSpiroStreamData       = NULL;
    s_MsmtGroupData *msmtGroupSpiroSummaryData      = NULL;
    s_MsmtGroupData *msmtGroupSpiroSessionData      = NULL;

    unsigned short refs[1];
    unsigned short session_id;
    unsigned short fvc_ids[3];   // For msmts that are used in the final summary
    int fvc_id_index = 0;
    unsigned short fev1_ids[3];
    int fev1_id_index = 0;

    unsigned short spiro_sequence = 0;      // First dump settings, then stream, then maneuver, then summary

    // Spiro base variables
    short fev05_index                               = -1;
    short fev075_index                              = -1;
    short fev1_index                                = -1;
    short fev3_index                                = -1;
    short fev6_index                                = -1;
    short fvc_index                                 = -1;
    short pef_index                                 = -1;
    short fef25_index                               = -1;
    short fef50_index                               = -1;
    short fef75_index                               = -1;
    short fef25_75_index                            = -1;
    short fet_index                                 = -1;
    short tpef_index                                = -1;
    short extrap_index                              = -1;
    short temp_index                                = -1;
    short humid_index                               = -1;
    short airPress_index                            = -1;
    short fev1Ratings_index                         = -1;
    // Spiro settings variables
    short age_index                                 = -1;
    short weight_index                              = -1;
    short height_index                              = -1;
    long sex_index                                  = -1;
    long ethnicity_index                            = -1;
    // Spiro session variables
    short session_index                             = -1;
    // Spiro streaming variables
    short flow_index                                = -1;
    short volume_index                              = -1;
    // Spiro summary variables
    short fev1AtsGrade_index                        = -1;
    short fvcAtsGrade_index                         = -1;
#endif

unsigned char *getBtAddress(void)
{
    return bluetoothAddress;
}

void configureSpecializations(void)
{
    bool result;
    unsigned char group_id = 1;

    s_SystemInfo *systemInfo = NULL;
    // Now to configure our GHS PHD
    // Create the time info data buffer
    #if (USES_TIMESTAMP == 1)
        createGhsTime(&sGhsTime, GHS_TIME_OFFSET_UNSUPPORTED, true, MDC_TIME_SYNC_NONE);
    #endif
    createTimeInfo(&sTimeInfo, sGhsTime, TIME_FLAGS_EPOCH_TIME, true);
    createCurrentTimeDataBuffer(&sTimeInfoData, sTimeInfo);

    // Create the system info data buffer
    result = createSystemInfo(&systemInfo, 1);
    result = addSpecialization(&systemInfo, SPECIALIZATION, 2);
    result = setSystemIdentifierByte(&systemInfo, systemId);
    result = setRequiredSystemInfoStrings(&systemInfo, MANUFACTURER_NAME, MODEL_NUMBER);
    result = setRegulationStatus(&systemInfo, true, false);
    result = setOptionalSystemInfoStrings(&systemInfo, "snX-4956", "fw90.9", "hw5.7.61", "sw0.0.5");
    result = createSystemInfoData(&systemInfoData, systemInfo);
    cleanUpSystemInfo(&systemInfo);

    #if (BP_CUFF == 1)
        memset(&storedBpMsmts, 0, NUMBER_OF_STORED_MSMTS * sizeof(s_MsmtDataBp));
        // Create the msmt data group 
        s_MsmtGroup *msmtGroup = NULL;
        s_GhsMsmt *bp = NULL;
        s_GhsMsmt *pr = NULL;
        s_Compound compounds[3];
        compounds[0].subType = MDC_PRESS_BLD_NONINV_SYS;
        compounds[1].subType = MDC_PRESS_BLD_NONINV_DIA;
        compounds[2].subType = MDC_PRESS_BLD_NONINV_MEAN;
        result = createMsmtGroup(&msmtGroup, true, 2, group_id++); // 2 msmts - bp, PR
        result = createCompoundNumericMsmt(&bp, MDC_PRESS_BLD_NONINV, true, MDC_DIM_MMHG, 3, compounds);
        unsigned long supplementalTypes[1] = {MDC_UPEXT_ARM_UPPER};
        result = setGhsMsmtSupplementalTypes(&bp, supplementalTypes, 1);
        bp_index = addGhsMsmtToGroup(bp, &msmtGroup);
        result = createNumericMsmt(&pr, MDC_PULS_RATE_NON_INV, true, MDC_DIM_BEAT_PER_MIN);
        pr_index = addGhsMsmtToGroup(pr, &msmtGroup);
        result = createMsmtGroupDataArray(&msmtGroupBpData, msmtGroup, sGhsTime, PACKET_TYPE_NORMAL);
        cleanUpMsmtGroup(&msmtGroup); // cleans up any allocated data -  we only need the data array now
    #endif  // BP cuff


    #if (PULSE_OX == 1)
        memset(&storedPoMsmts, 0, NUMBER_OF_STORED_MSMTS * sizeof(s_MsmtDataPo));
        // Create the msmt data group for SPOT measurements
        s_MsmtGroup *msmtGroup = NULL;
        s_GhsMsmt *spo2 = NULL;
        s_GhsMsmt *pr = NULL;
        s_GhsMsmt *qual = NULL;
        result = createMsmtGroup(&msmtGroup, true, 3, group_id++); // 3 msmts - SpO2, PR, Qual
        unsigned long supplementalTypes[1] = {MDC_MODALITY_SPOT};
        result = setHeaderSupplementalTypes(&msmtGroup, supplementalTypes, 1);
        result = createNumericMsmt(&spo2, MDC_PULS_OXIM_SAT_O2, true, MDC_DIM_PERCENT);
        spo2_index = addGhsMsmtToGroup(spo2, &msmtGroup);
        result = createNumericMsmt(&pr, MDC_PULS_OXIM_PULS_RATE, true, MDC_DIM_BEAT_PER_MIN);
        pr_index = addGhsMsmtToGroup(pr, &msmtGroup);
        result = createNumericMsmt(&qual, MDC_SAT_O2_QUAL, true, MDC_DIM_PERCENT);
        qual_index = addGhsMsmtToGroup(qual, &msmtGroup);
        result = createMsmtGroupDataArray(&msmtGroupSpotData, msmtGroup, sGhsTime, PACKET_TYPE_NORMAL);
        cleanUpMsmtGroup(&msmtGroup); // cleans up any allocated data -  we only need the data array now

        spo2 = NULL;
        pr = NULL;
        qual = NULL;

        // Now for the continuous
        result = createMsmtGroup(&msmtGroup, false, 3, group_id); // 3 msmts - SpO2, PR, Qual
        result = createNumericMsmt(&spo2, MDC_PULS_OXIM_SAT_O2, true, MDC_DIM_PERCENT);
        spo2_cont_index = addGhsMsmtToGroup(spo2, &msmtGroup);
        result = createNumericMsmt(&pr, MDC_PULS_OXIM_PULS_RATE, true, MDC_DIM_BEAT_PER_MIN);
        pr_cont_index = addGhsMsmtToGroup(pr, &msmtGroup);
        result = createNumericMsmt(&qual, MDC_SAT_O2_QUAL, true, MDC_DIM_PERCENT);
        qual_cont_index = addGhsMsmtToGroup(qual, &msmtGroup);
        #if (SEND_OPTIMIZED == 1)
            result = createMsmtGroupDataArray(&msmtGroupContData, msmtGroup, NULL, PACKET_TYPE_OPTIMIZED_FIRST);
            result = createMsmtGroupDataArray(&msmtGroupOptimizedContData, msmtGroup, NULL, PACKET_TYPE_OPTIMIZED_FOLLOWS);
        #else
            result = createMsmtGroupDataArray(&msmtGroupContData, msmtGroup, NULL, PACKET_TYPE_NORMAL);
        #endif
        cleanUpMsmtGroup(&msmtGroup); // cleans up any allocated data - we only need the data array now
    #endif  // Pulse ox

    #if (SPIROMETER == 1)
        spiro_sequence = 0;
        // Going to lad all the data here and send it in chunks on the 'live' command and then stop.
        // The only update will be the time stamps
        s_MsmtGroup *spiroSettingsGroup = NULL;
        s_GhsMsmt *age = NULL;
        s_GhsMsmt *weight = NULL;
        s_GhsMsmt *height = NULL;
        s_GhsMsmt *sex = NULL;
        s_GhsMsmt *ethnicity = NULL;
        s_MsmtGroup *spiroSessionGroup = NULL;
        s_GhsMsmt *session = NULL;
        s_MsmtGroup *spiroStreamingGroup = NULL;
        s_GhsMsmt *volume = NULL;
        s_GhsMsmt *flow = NULL;
        s_MsmtGroup *spiroManeuvGroup = NULL;
        s_GhsMsmt *fev05 = NULL;
        s_GhsMsmt *fev075 = NULL;
        s_GhsMsmt *fev1 = NULL;
        s_GhsMsmt *fev3 = NULL;
        s_GhsMsmt *fev6 = NULL;
        s_GhsMsmt *fvc = NULL;
        s_GhsMsmt *pef = NULL;
        s_GhsMsmt *fef25 = NULL;
        s_GhsMsmt *fef50 = NULL;
        s_GhsMsmt *fef75 = NULL;
        s_GhsMsmt *fef25_75 = NULL;
        s_GhsMsmt *fet = NULL;
        s_GhsMsmt *tpef = NULL;
        s_GhsMsmt *extrap = NULL;
        s_GhsMsmt *temp = NULL;
        s_GhsMsmt *humid = NULL;
        s_GhsMsmt *airPress = NULL;
        s_GhsMsmt *fev1Ratings = NULL;
        s_MsmtGroup *spiroSummaryGroup = NULL;
        s_GhsMsmt *fvcAtsGrade = NULL;
        s_GhsMsmt *fev1AtsGrade = NULL;

        // settings
        result = createMsmtGroup(&spiroSettingsGroup, false, 3, group_id++); // 5 msmts
        result = setHeaderOptions(&spiroSettingsGroup, true, false, 0);
        result = createNumericMsmt(&age, MDC_HF_AGE, true, MDC_DIM_YR);
        result = createNumericMsmt(&weight, MDC_MASS_BODY_ACTUAL, false, MDC_DIM_KILO_G);
        result = createNumericMsmt(&height, MDC_LEN_BODY_ACTUAL, false, MDC_DIM_CENTI_M);
        age_index = addGhsMsmtToGroup(age, &spiroSettingsGroup);
        weight_index = addGhsMsmtToGroup(weight, &spiroSettingsGroup);
        height_index = addGhsMsmtToGroup(height, &spiroSettingsGroup);
        result = createMsmtGroupDataArray(&msmtGroupSpiroSettingsData, spiroSettingsGroup, NULL, PACKET_TYPE_NORMAL);
        s_MderFloat mder;
        mder.exponent = 0;
        mder.mantissa = AGE;
        mder.mderFloatType = MDER_SFLOAT;
        mder.specialValue = MDER_NUMBER;
        updateDataNumeric(&msmtGroupSpiroSettingsData, age_index, &mder, msmt_id++);
        mder.exponent = -2;
        mder.mantissa = WEIGHT;
        mder.mderFloatType = MDER_FLOAT;
        updateDataNumeric(&msmtGroupSpiroSettingsData, weight_index, &mder, msmt_id++);
        mder.exponent = -1;
        mder.mantissa = HEIGHT;
        mder.mderFloatType = MDER_FLOAT;
        updateDataNumeric(&msmtGroupSpiroSettingsData, height_index, &mder, msmt_id++);
        cleanUpMsmtGroup(&spiroSettingsGroup);

        // session
        result = createMsmtGroup(&spiroSessionGroup, true, 1, group_id++); // 1 msmt
        result = createCodedMsmt(&session, MDC_DIAG_SESSION);
        session_index = addGhsMsmtToGroup(session, &spiroSessionGroup);
        createMsmtGroupDataArray(&msmtGroupSpiroSessionData, spiroSessionGroup, sGhsTime, PACKET_TYPE_NORMAL);
        cleanUpMsmtGroup(&spiroSessionGroup);
        refs[0] = session_id;

        // streaming data
        
        result = createMsmtGroup(&spiroStreamingGroup, true, 1, group_id++); // 2 msmts
        setHeaderSrcHandleRefs(&spiroStreamingGroup, 1 );
        s_MderFloat period;
        period.exponent = -3;
        period.mantissa = 10;  // 10 milliseconds; 0.01 seconds
        period.mderFloatType = MDER_FLOAT;
        period.specialValue = MDER_NUMBER;
        s_MderFloat scaleFactor;
        scaleFactor.exponent = 0;
        scaleFactor.mantissa = 1;  // raw data will be milliliters
        scaleFactor.mderFloatType = MDER_FLOAT;
        scaleFactor.specialValue = MDER_NUMBER;
        s_MderFloat offset;
        offset.exponent = 0;
        offset.mantissa = 0;  // raw data will be milliliters
        offset.mderFloatType = MDER_FLOAT;
        offset.specialValue = MDER_NUMBER;
      //  result = createRtsaMsmt(&volume, MDC_VOL_AWAY, MDC_DIM_MILLI_L, &period, &scaleFactor, &offset, NO_OF_SAMPLES, SAMPLE_SIZE);
        result = createRtsaMsmt(&flow, MDC_FLOW_AWAY, MDC_DIM_MILLI_L_PER_SEC, &period, &scaleFactor, &offset, NO_OF_SAMPLES, SAMPLE_SIZE);
      //  volume_index = addGhsMsmtToGroup(volume, &spiroStreamingGroup);
        flow_index = addGhsMsmtToGroup(flow, &spiroStreamingGroup);
        createMsmtGroupDataArray(&msmtGroupSpiroStreamData, spiroStreamingGroup, sGhsTime, PACKET_TYPE_NORMAL);
        cleanUpMsmtGroup(&spiroStreamingGroup);


        // maneuver
        result = createMsmtGroup(&spiroManeuvGroup, true, 6, group_id++); // 6 msmts for now
        setHeaderSrcHandleRefs(&spiroManeuvGroup, 1);
        result = createNumericMsmt(&fev1, MDC_VOL_AWAY_EXP_FORCED_1S, false, MDC_DIM_L);
        result = createNumericMsmt(&fev6, MDC_VOL_AWAY_EXP_FORCED_6S, false, MDC_DIM_L);
        result = createNumericMsmt(&fvc, MDC_VOL_AWAY_EXP_FORCED_CAPACITY, false, MDC_DIM_L);
        result = createNumericMsmt(&pef, MDC_FLOW_AWAY_EXP_FORCED_PEAK, false, MDC_DIM_L_PER_SEC);
        result = createNumericMsmt(&fet, MDC_VOL_AWAY_EXP_FORCED_TIME, false, MDC_DIM_SEC);
        s_Compound compounds[3];
        compounds[0].subType = MDC_VOL_AWAY_FEV1_Z_SCORE;
        compounds[0].subUnits = MDC_DIM_DIMLESS;
        compounds[1].subType = MDC_VOL_AWAY_FEV1_LLN;
        compounds[1].subUnits = MDC_DIM_L;
        compounds[2].subType = MDC_VOL_AWAY_FEV1_PERCENT_PRED;
        compounds[2].subUnits = MDC_DIM_PERCENT;
        result = createComplexCompoundNumericMsmt(&fev1Ratings, MDC_VOL_AWAY_FEV1_RATINGS, false, 3, compounds);
        unsigned long supplementalTypes[3];
        supplementalTypes[0] = MDC_SPIRO_PRED_EQN_NHANESIII;
        supplementalTypes[1] = MDC_ETHNICITY_WHITE;
        supplementalTypes[2] = MDC_SEX_AT_BIRTH_MALE;
        setGhsMsmtSupplementalTypes(&fev1Ratings, supplementalTypes, 3);
        fev1_index = addGhsMsmtToGroup(fev1, &spiroManeuvGroup);
        fev6_index = addGhsMsmtToGroup(fev6, &spiroManeuvGroup);
        fvc_index = addGhsMsmtToGroup(fvc, &spiroManeuvGroup);
        pef_index = addGhsMsmtToGroup(pef, &spiroManeuvGroup);
        fet_index = addGhsMsmtToGroup(fet, &spiroManeuvGroup);
        fev1Ratings_index = addGhsMsmtToGroup(fev1Ratings, &spiroManeuvGroup);
        setGhsMsmtSrcHandleRefs(&fev1Ratings, 1);
        createMsmtGroupDataArray(&msmtGroupSpiroManeuvData, spiroManeuvGroup, sGhsTime, PACKET_TYPE_NORMAL);
        cleanUpMsmtGroup(&spiroManeuvGroup);

        // summary
        // It makes not sense to create this data array here as the number of references will be unknown until
        // the session ends. Then we pick the best 3 best FVC and FEV1 values and use that. Then we fill in
        // the refs which ideally will have up to three entries in addition to the session_id.
        result = createMsmtGroup(&spiroSummaryGroup, true, 2, group_id++); // 1 msmts
        setHeaderSrcHandleRefs(&spiroSummaryGroup, 1);
        result = createCodedMsmt(&fvcAtsGrade, MDC_SPIRO_FVC_ATS_QUAL);
        setGhsMsmtSrcHandleRefs(&fvcAtsGrade, 2);
        result = createCodedMsmt(&fev1AtsGrade, MDC_SPIRO_FEV1_ATS_QUAL);
        setGhsMsmtSrcHandleRefs(&fev1AtsGrade, 2);
        fvcAtsGrade_index = addGhsMsmtToGroup(fvcAtsGrade, &spiroSummaryGroup);
        fev1AtsGrade_index = addGhsMsmtToGroup(fev1AtsGrade, &spiroSummaryGroup);
        createMsmtGroupDataArray(&msmtGroupSpiroSummaryData, spiroSummaryGroup, sGhsTime, PACKET_TYPE_NORMAL);
        cleanUpMsmtGroup(&spiroSummaryGroup);
    #endif
}

bool encodeSpecializationMsmts(void *data)
{
    #if (BP_CUFF)
        s_MsmtDataBp *bp = (s_MsmtDataBp *)data;
        s_MderFloat mder[3];
        if (!prepareMeasurements(msmtGroupBpData)) return false;
        // Add the command
        msmtGroupBpData->data[0] = (global_send.current_command & 0xFF);
        msmtGroupBpData->data[1] = ((global_send.current_command >> 8) & 0xFF);
        mder[0].mderFloatType = MDER_SFLOAT;
        mder[0].specialValue = MDER_NUMBER;
        mder[0].exponent = 0;
        mder[0].mantissa = bp->systolic;
        mder[1].mderFloatType = MDER_SFLOAT;
        mder[1].specialValue = MDER_NUMBER;
        mder[1].exponent = 0;
        mder[1].mantissa = bp->diastolic;
        mder[2].mderFloatType = MDER_SFLOAT;
        mder[2].specialValue = MDER_NUMBER;
        mder[2].exponent = 0;
        mder[2].mantissa = bp->mean;
        updateDataCompound(&msmtGroupBpData, bp_index, mder, msmt_id++);
        mder[0].mantissa = bp->pulseRate;
        updateDataNumeric(&msmtGroupBpData, pr_index, &mder[0], msmt_id++);
        updateTimeStampEpoch(&msmtGroupBpData, bp->sGhsTime.epoch);
        NRF_LOG_DEBUG("Bp msmt to send");
    #endif
    #if (PULSE_OX == 1)
        s_MsmtDataPo *po = (s_MsmtDataPo *)data;
        s_MderFloat mder;
        mder.mderFloatType = MDER_SFLOAT;
        mder.specialValue = MDER_NUMBER;
        if (po->isContinuous)
        {
            #if (SEND_OPTIMIZED == 1)
                s_MsmtGroupData *bytes = first_cont_sent ? msmtGroupOptimizedContData : msmtGroupContData;
            #else
                s_MsmtGroupData *bytes = msmtGroupContData;
            #endif
            if (!prepareMeasurements(bytes)) return false;
            // Add the command
            bytes->data[0] = (global_send.current_command & 0xFF);
            bytes->data[1] = ((global_send.current_command >> 8) & 0xFF);
            mder.exponent = 0;
            mder.mantissa = po->spo2;
            updateDataNumeric(&bytes, spo2_cont_index, &mder, msmt_id++);
            mder.mantissa = po->pulseRate;
            updateDataNumeric(&bytes, pr_cont_index, &mder, msmt_id++);
            mder.mantissa = po->pulseQuality;
            mder.exponent = -2;
            updateDataNumeric(&bytes, qual_cont_index, &mder, msmt_id++);
            NRF_LOG_DEBUG("Continuous msmt to send");
        }
        else
        {
            if (!prepareMeasurements(msmtGroupSpotData)) return false;
            // Add the command
            msmtGroupSpotData->data[0] = (global_send.current_command & 0xFF);
            msmtGroupSpotData->data[1] = ((global_send.current_command >> 8) & 0xFF);

            mder.exponent = 0;
            mder.mantissa = po->spo2;
            updateDataNumeric(&msmtGroupSpotData, spo2_index, &mder, msmt_id++);
            mder.mantissa = po->pulseRate;
            updateDataNumeric(&msmtGroupSpotData, pr_index, &mder, msmt_id++);
            mder.mantissa = po->pulseQuality;
            mder.exponent = -2;
            updateDataNumeric(&msmtGroupSpotData, qual_index, &mder, msmt_id++);
            NRF_LOG_DEBUG("Spot msmt to send");
            updateTimeStampEpoch(&msmtGroupSpotData, po->sGhsTime.epoch);
        }
    #endif
    // We are not generating the measurements on the fly as in the other cases, it is all pre done
    // except for the time stamps.
    #if (SPIROMETER == 1)
        s_SpiroSession *session = (s_SpiroSession *)data;
        if (spiro_sequence == 0)
        {
            if (!prepareMeasurements(msmtGroupSpiroSettingsData)) return false;
            // Add the command
            msmtGroupSpiroSettingsData->data[0] = (global_send.current_command & 0xFF);
            msmtGroupSpiroSettingsData->data[1] = ((global_send.current_command >> 8) & 0xFF);
            NRF_LOG_DEBUG("Sending Spirometer settings");
        }
        else if (spiro_sequence == 1)
        {
            if (!prepareMeasurements(msmtGroupSpiroSessionData)) return false;
            // Add the command
            msmtGroupSpiroSessionData->data[0] = (global_send.current_command & 0xFF);
            msmtGroupSpiroSessionData->data[1] = ((global_send.current_command >> 8) & 0xFF);
            NRF_LOG_DEBUG("Sending Spirometer session");
            updateTimeStampEpoch(&msmtGroupSpiroSessionData, session->sGhsTime.epoch);
            session_id = msmt_id;
            updateDataCoded(&msmtGroupSpiroSessionData, session_index, MDC_LUNG_DIAG_SPIRO, msmt_id++);
        }
        else if (spiro_sequence == 2)
        {
            NRF_LOG_DEBUG("Sending Spirometer flow and volume streams 1");
            if (!prepareMeasurements(msmtGroupSpiroStreamData)) return false;
            // Add the command
            msmtGroupSpiroStreamData->data[0] = (global_send.current_command & 0xFF);
            msmtGroupSpiroStreamData->data[1] = ((global_send.current_command >> 8) & 0xFF);
            updateTimeStampEpoch(&msmtGroupSpiroStreamData, session->sGhsTime.epoch);
            updateDataHeaderSrcHandleRef(&msmtGroupSpiroStreamData, session_id, 0);
            updateDataRtsa(&msmtGroupSpiroStreamData, flow_index, flowBytes, NO_OF_SAMPLES * SAMPLE_SIZE, msmt_id++);
        }
        else if (spiro_sequence == 3)
        {
            NRF_LOG_DEBUG("Sending Spirometer flow and volume streams 2");
            if (!prepareMeasurements(msmtGroupSpiroStreamData)) return false;
            // Add the command
            msmtGroupSpiroStreamData->data[0] = (global_send.current_command & 0xFF);
            msmtGroupSpiroStreamData->data[1] = ((global_send.current_command >> 8) & 0xFF);
            updateTimeStampEpoch(&msmtGroupSpiroStreamData, session->sGhsTime.epoch);
            updateDataRtsa(&msmtGroupSpiroStreamData, flow_index, &flowBytes[NO_OF_SAMPLES], NO_OF_SAMPLES * SAMPLE_SIZE, msmt_id++);
        }
        else if (spiro_sequence == 4)
        {
            if (!prepareMeasurements(msmtGroupSpiroManeuvData)) return false;
            // Add the command
            msmtGroupSpiroManeuvData->data[0] = (global_send.current_command & 0xFF);
            msmtGroupSpiroManeuvData->data[1] = ((global_send.current_command >> 8) & 0xFF);
            NRF_LOG_DEBUG("Sending Spirometer Maneuver data");
            updateTimeStampEpoch(&msmtGroupSpiroManeuvData, session->sGhsTime.epoch);
            updateDataHeaderSrcHandleRef(&msmtGroupSpiroManeuvData, session_id, 0);
            s_MderFloat mder;
            mder.exponent = -3;
            mder.mantissa = fev1Val;
            mder.mderFloatType = MDER_FLOAT;
            mder.specialValue = MDER_NUMBER;
            fev1_ids[fev1_id_index] = msmt_id;
            updateDataNumeric(&msmtGroupSpiroManeuvData, fev1_index, &mder, msmt_id++);
            mder.mantissa = fev6Val;
            updateDataNumeric(&msmtGroupSpiroManeuvData, fev6_index, &mder, msmt_id++);
            mder.mantissa = fvcVal;
            fvc_ids[fvc_id_index] = msmt_id;
            updateDataNumeric(&msmtGroupSpiroManeuvData, fvc_index, &mder, msmt_id++);
            mder.mantissa = pefVal;
            updateDataNumeric(&msmtGroupSpiroManeuvData, pef_index, &mder, msmt_id++);
            mder.mantissa = fetVal;
            updateDataNumeric(&msmtGroupSpiroManeuvData, fet_index, &mder, msmt_id++);
            s_MderFloat mders[3];
            mders[0].exponent = -3;
            mders[0].mantissa = fev1RatingsVal[0];       // Z score of +LLN value
            mders[0].mderFloatType = MDER_FLOAT;
            mders[0].specialValue = MDER_NUMBER;
            mders[1].exponent = -3;
            mders[1].mantissa = fev1RatingsVal[1];
            mders[1].mderFloatType = MDER_FLOAT;
            mders[1].specialValue = MDER_NUMBER;
            mders[2].exponent = -1;
            mders[2].mantissa = fev1RatingsVal[2]; // 100 * 10 * fev1 / fev1 predicited;
            mders[2].mderFloatType = MDER_FLOAT;
            mders[2].specialValue = MDER_NUMBER;
            updateDataCompound(&msmtGroupSpiroManeuvData, fev1Ratings_index, &mders[0], msmt_id++);
            updateDataGhsMsmtSrcHandleRef(&msmtGroupSpiroManeuvData, fev1Ratings_index, fev1_ids[fev1_id_index], 0);
            fev1_id_index++;
            fvc_id_index++;
        }
        else if (spiro_sequence == 5)
        {
            if (!prepareMeasurements(msmtGroupSpiroSummaryData)) return false;
            // Add the command
            msmtGroupSpiroSummaryData->data[0] = (global_send.current_command & 0xFF);
            msmtGroupSpiroSummaryData->data[1] = ((global_send.current_command >> 8) & 0xFF);
            NRF_LOG_DEBUG("Sending Spirometer summary data");
            updateTimeStampEpoch(&msmtGroupSpiroSummaryData, session->sGhsTime.epoch);
            updateDataHeaderSrcHandleRef(&msmtGroupSpiroSummaryData, session_id, 0);
            updateDataCoded(&msmtGroupSpiroSummaryData, fvcAtsGrade_index, fvcAtsGradeVal, msmt_id++);
            updateDataCoded(&msmtGroupSpiroSummaryData, fev1AtsGrade_index, fev1AtsGradeVal, msmt_id++);
            updateDataGhsMsmtSrcHandleRef(&msmtGroupSpiroSummaryData, fvcAtsGrade_index, fvc_ids[0], 0);
            updateDataGhsMsmtSrcHandleRef(&msmtGroupSpiroSummaryData, fvcAtsGrade_index, fev1_ids[0], 1);
            updateDataGhsMsmtSrcHandleRef(&msmtGroupSpiroSummaryData, fev1AtsGrade_index, fev1_ids[0], 0);
            updateDataGhsMsmtSrcHandleRef(&msmtGroupSpiroSummaryData, fev1AtsGrade_index, fvc_ids[0], 1);
        }
        spiro_sequence++;
    #endif
    return true;
}

void generateAndAddStoredMsmt(unsigned long long timeStamp, unsigned short numberOfStoredMsmtGroups)
{
    #if (BP_CUFF == 1)

        storedBpMsmts[numberOfStoredMsmtGroups].hasTimeStamp = true;
        storedBpMsmts[numberOfStoredMsmtGroups].sGhsTime.epoch = epoch + timeStamp;
       // storedBpMsmts[numberOfStoredMsmtGroups].sGhsTime.dstShift = sGhsTime->dstShift;
        storedBpMsmts[numberOfStoredMsmtGroups].sGhsTime.offsetShift = sGhsTime->offsetShift;
        storedBpMsmts[numberOfStoredMsmtGroups].sGhsTime.timeSync = sGhsTime->timeSync;
        storedBpMsmts[numberOfStoredMsmtGroups].systolic = 95 + (timeStamp & 0x0F);
        storedBpMsmts[numberOfStoredMsmtGroups].diastolic = 55 + (timeStamp & 0x0F);
        storedBpMsmts[numberOfStoredMsmtGroups].mean = 
            (storedBpMsmts[numberOfStoredMsmtGroups].systolic +
             storedBpMsmts[numberOfStoredMsmtGroups].diastolic) / 2;
        storedBpMsmts[numberOfStoredMsmtGroups].pulseRate = 40 + (timeStamp & 0x07);
            NRF_LOG_DEBUG("Measurement added: sys %u, dia %u, mean %u, PR %u, timestamp %lu", 
                storedBpMsmts[numberOfStoredMsmtGroups].systolic,
                storedBpMsmts[numberOfStoredMsmtGroups].diastolic,
                storedBpMsmts[numberOfStoredMsmtGroups].mean,
                storedBpMsmts[numberOfStoredMsmtGroups].pulseRate, 
                timeStamp);
    #endif
    #if (PULSE_OX == 1)
        storedPoMsmts[numberOfStoredMsmtGroups].hasTimeStamp = true;
        storedPoMsmts[numberOfStoredMsmtGroups].sGhsTime.epoch = epoch + timeStamp;
     //   storedPoMsmts[numberOfStoredMsmtGroups].sGhsTime.dstShift = sGhsTime->dstShift;
        storedPoMsmts[numberOfStoredMsmtGroups].sGhsTime.offsetShift = sGhsTime->offsetShift;
        storedPoMsmts[numberOfStoredMsmtGroups].sGhsTime.timeSync = sGhsTime->timeSync;

        storedPoMsmts[numberOfStoredMsmtGroups].spo2 = 95 + (timeStamp & 0x03);
        storedPoMsmts[numberOfStoredMsmtGroups].pulseRate = 45 + (timeStamp & 0x07);
        storedPoMsmts[numberOfStoredMsmtGroups].pulseQuality = 523 + (timeStamp & 0xFF);

        NRF_LOG_DEBUG("Measurement added: SpO2 %u%, PR %u, Pulsatile X 100 %u%, timestamp %lu", 
            storedPoMsmts[numberOfStoredMsmtGroups].spo2, 
            storedPoMsmts[numberOfStoredMsmtGroups].pulseRate, 
            storedPoMsmts[numberOfStoredMsmtGroups].pulseQuality, timeStamp);
    #endif
    #if (SPIROMETER == 1)
        return;
    #endif
}

void generateLiveDataForSpecializations(unsigned long live_data_count, unsigned long long timeStamp)
{
    #if (BP_CUFF == 1)
        if ((live_data_count & 0x0F) == 0x00)
        {
            s_MsmtDataBp bpMsmt;
            memset(&bpMsmt, 0, sizeof(s_MsmtDataBp));
            bpMsmt.systolic = 100 + (timeStamp & 0x0F);
            bpMsmt.diastolic = 55 + (timeStamp & 0x0F);
            bpMsmt.mean = ((bpMsmt.systolic + bpMsmt.diastolic) >> 1);
            bpMsmt.pulseRate = 40 + (timeStamp & 0x07);
            bpMsmt.hasTimeStamp = true;
            bpMsmt.sGhsTime.epoch = epoch + timeStamp;
        //    bpMsmt.sGhsTime.dstShift = sGhsTime->dstShift;
            bpMsmt.sGhsTime.offsetShift = sGhsTime->offsetShift;
            bpMsmt.sGhsTime.timeSync = sGhsTime->timeSync;
            NRF_LOG_DEBUG("Measurement added to queue: sys %u, dia %u, mean %u, PR %u", 
                bpMsmt.systolic,
                bpMsmt.diastolic,
                bpMsmt.mean,
                bpMsmt.pulseRate);
            if(sd_mutex_acquire(&q_mutex) != NRF_ERROR_SOC_MUTEX_ALREADY_TAKEN)
            {
                enqueue(queue, &bpMsmt, sizeof(s_MsmtDataBp));
                sd_mutex_release(&q_mutex);
            }
        }
        else
        {
            return;
        }
    #endif
    
    #if (PULSE_OX == 1)
        s_MsmtDataPo poMsmt;
        memset(&poMsmt, 0, sizeof(s_MsmtDataPo));
        poMsmt.spo2 = 95 + (timeStamp & 0x03);
        poMsmt.pulseRate = 45 + (timeStamp & 0x07);
        poMsmt.pulseQuality = 523 + (timeStamp & 0xFF);
        if ((live_data_count & 0x0F) == 0x00)
        {
            poMsmt.hasTimeStamp = true;
            poMsmt.sGhsTime.epoch = epoch + timeStamp;
        //    poMsmt.sGhsTime.dstShift = sGhsTime->dstShift;
            poMsmt.sGhsTime.offsetShift = sGhsTime->offsetShift;
            poMsmt.sGhsTime.timeSync = sGhsTime->timeSync;
            NRF_LOG_DEBUG("Measurement added to queue: spo2 %u, pulse rate %u, pulse quality %u", 
                poMsmt.spo2,
                poMsmt.pulseRate,
                poMsmt.pulseQuality);
            poMsmt.isContinuous = false;
        }
        else
        {
            poMsmt.hasTimeStamp = false;
            poMsmt.isContinuous = true;
        }
        if(sd_mutex_acquire(&q_mutex) != NRF_ERROR_SOC_MUTEX_ALREADY_TAKEN)
        {
            enqueue(queue, &poMsmt, sizeof(s_MsmtDataPo));
            sd_mutex_release(&q_mutex);
        }
    #endif
    #if (SPIROMETER == 1)
        if (spiro_sequence >= 6) return;
        s_SpiroSession session;  // Dummy - data is already encoded
        memset(&session, 0, sizeof(s_SpiroSession));
        session.sGhsTime.epoch = epoch + timeStamp;
        session.sGhsTime.offsetShift = sGhsTime->offsetShift;
        session.sGhsTime.timeSync = sGhsTime->timeSync;
        if(sd_mutex_acquire(&q_mutex) != NRF_ERROR_SOC_MUTEX_ALREADY_TAKEN)
        {
            enqueue(queue, &session, sizeof(s_SpiroSession));
            sd_mutex_release(&q_mutex);
        }
    #endif
}

void handleSpecializationsOnSetTime(unsigned short numberOfStoredMsmtGroups, long long diff, unsigned short timeSync)
{
    int i;
    #if (BP_CUFF == 1)
        updateTimeStampTimeSync(&msmtGroupBpData, timeSync);
    #endif
    #if (PULSE_OX == 1)
        updateTimeStampTimeSync(&msmtGroupSpotData, timeSync);
    #endif
    for (i = 0; i < numberOfStoredMsmtGroups; i++)
    {
    //TODO: Check that the measurements are on the current time line, if not, leave alone.
        #if (PULSE_OX == 1)
            storedPoMsmts[i].sGhsTime.epoch = storedPoMsmts[i].sGhsTime.epoch + diff;
            storedPoMsmts[i].sGhsTime.timeSync = timeSync;
        #endif
        #if (BP_CUFF == 1)
             storedBpMsmts[i].sGhsTime.epoch = storedBpMsmts[i].sGhsTime.epoch + diff;
             storedBpMsmts[i].sGhsTime.timeSync = timeSync;
        #endif
    }
}

void sendStoredSpecializationMsmts(unsigned short stored_count)
{
    // Set up parameters for notification of this PDU - likely in fragments
    #if (BP_CUFF == 1)
        NRF_LOG_DEBUG("Stored Measurements added to queue: sys %u, dia %u, mean %u, PR %u, count %d", 
            storedBpMsmts[stored_count].systolic,
            storedBpMsmts[stored_count].diastolic,
            storedBpMsmts[stored_count].mean,
            storedBpMsmts[stored_count].pulseRate,
            stored_count);
        if(sd_mutex_acquire(&q_mutex) != NRF_ERROR_SOC_MUTEX_ALREADY_TAKEN)
        {
            enqueue(queue, &storedBpMsmts[stored_count], sizeof(s_MsmtDataBp));
            sd_mutex_release(&q_mutex);
        }
    #endif
    #if (PULSE_OX == 1)
        NRF_LOG_DEBUG("Stored Measurements added to queue: spo2 %u, pr %u, qual %u, count %d", 
            storedPoMsmts[stored_count].spo2,
            storedPoMsmts[stored_count].pulseRate,
            storedPoMsmts[stored_count].pulseQuality,
            stored_count);
        if(sd_mutex_acquire(&q_mutex) != NRF_ERROR_SOC_MUTEX_ALREADY_TAKEN)
        {
            enqueue(queue, &storedPoMsmts[stored_count], sizeof(s_MsmtDataPo));
            sd_mutex_release(&q_mutex);
        }
    #endif
}

void deleteStoredSpecializationMsmts(void)
{
    #if (PULSE_OX == 1)
        memset(&storedPoMsmts, 0, NUMBER_OF_STORED_MSMTS * sizeof(s_MsmtDataPo));
    #endif
    #if (BP_CUFF == 1)
        memset(&storedBpMsmts, 0, NUMBER_OF_STORED_MSMTS * sizeof(s_MsmtDataBp));
    #endif
}

void cleanUpSpecializations(void)
{
    #if (BP_CUFF == 1)
        cleanUpMsmtGroupData(&msmtGroupBpData);
    #endif

    #if (PULSE_OX == 1)
        cleanUpMsmtGroupData(&msmtGroupSpotData);
        cleanUpMsmtGroupData(&msmtGroupContData);
        cleanUpMsmtGroupData(&msmtGroupOptimizedContData);
    #endif
    #if (SPIROMETER == 1)
        cleanUpMsmtGroupData(&msmtGroupSpiroSummaryData);
        cleanUpMsmtGroupData(&msmtGroupSpiroManeuvData);
        cleanUpMsmtGroupData(&msmtGroupSpiroStreamData);
        cleanUpMsmtGroupData(&msmtGroupSpiroSessionData);
        cleanUpMsmtGroupData(&msmtGroupSpiroSettingsData);
    #endif
    cleanUpGhsTime(&sGhsTime);
    cleanUpTimeInfoData(&sTimeInfoData);
    cleanUpTimeInfo(&sTimeInfo);
    cleanUpSystemInfoData(&systemInfoData);
}